@@ -62,12 +62,10 @@ class InputParameters(object):
         self.cheb_ang_order = 6
         self.cheb_ang_cutoff = 4.0
         self.cheb_delta = 1e-4
-        self.cheb_n_jobs = -1
         self.cheb_param = {'rad_order': self.cheb_rad_order,
                            'rad_cutoff': self.cheb_rad_cutoff,
                            'ang_order': self.cheb_ang_order,
-                           'ang_cutoff': self.self.cheb_ang_cutoff,
-                           'cheb_n_jobs': self.cheb_n_jobs,
+                           'ang_cutoff': self.cheb_ang_cutoff,
                            'delta': self.cheb_delta}
 
         self.kerneltype        = 'sqexp'
