@@ -9,14 +9,13 @@ from aenet_gpr.src.pytorch_kernel import FPKernel
 from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch
 
 
-def chebyshev_descriptor_gradient(atoms, model, atoms_mask, delta=1e-4, n_jobs=-1, dtype='float32'):
+def chebyshev_descriptor_gradient(atoms, model, atoms_mask, delta=1e-4, dtype=torch.float64):
     """
     Args:
         atoms (ase.Atoms):
         model: Chebyshev calculator
         atoms_mask: Atom index to keep
         delta (float): displacement (Ang)
-        n_jobs (int): number of processes
         dtype (dtype):
 
     Returns:
@@ -24,120 +23,129 @@ def chebyshev_descriptor_gradient(atoms, model, atoms_mask, delta=1e-4, n_jobs=-
         grad (torch.Tensor): (n_reduced_atoms, n_reduced_atoms, 3, descriptor_dim)
     """
     # Get original descriptor
-    atoms_mask = atoms_mask.cpu().numpy()
-    if np.any(atoms.pbc):
-        desc = model.featurize_structure(atoms.positions,
-                                         atoms.get_chemical_symbols(),
-                                         cell=np.array([atoms.cell[i] for i in range(3)]),
-                                         pbc=atoms.pbc)
-        n_atoms, D = desc.shape
+    batch_positions = [(torch.tensor(atoms.positions, dtype=dtype))]
+    original_positions = atoms.get_positions()
 
-        # Pre-allocate gradient array
-        grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+    for i in atoms_mask:
+        for j in range(3):
+            # Forward perturbation
+            pos_f = original_positions.copy()
+            pos_f[i, j] += delta
+            batch_positions.append(torch.tensor(pos_f, dtype=dtype))
 
-        # Get original positions once
-        original_positions = atoms.get_positions()
+            # Backward perturbation
+            pos_b = original_positions.copy()
+            pos_b[i, j] -= delta
+            batch_positions.append(torch.tensor(pos_b, dtype=dtype))
 
-        # Pre-create all perturbations
-        perturbations = []
-        for i in atoms_mask:
-            for j in range(3):  # x, y, z
-                # Forward perturbation
-                pos_f = original_positions.copy()
-                pos_f[i, j] += delta
-                perturbations.append(('f', i, j, pos_f))
+    batch_species = [atoms.get_chemical_symbols() for i in range(len(batch_positions))]
 
-                # Backward perturbation
-                pos_b = original_positions.copy()
-                pos_b[i, j] -= delta
-                perturbations.append(('b', i, j, pos_b))
+    # (Natoms * (6 * Nmaks + 1), Ndescriptor)
+    features_batch, batch_indices = model(batch_positions,
+                                          batch_species)
 
-        if n_jobs == 1:
-            # Compute all descriptors
-            all_descriptors = []
-            for direction, i, j, positions in perturbations:
-                atoms_temp = deepcopy(atoms)
-                atoms_temp.set_positions(positions)
-                desc_temp = model.featurize_structure(atoms.positions,
-                                                      atoms.get_chemical_symbols(),
-                                                      cell=np.array([atoms.cell[i] for i in range(3)]),
-                                                      pbc=atoms.pbc)
-                all_descriptors.append((direction, i, j, desc_temp))
-        else:
-            # --- Parallelize descriptor computation ---
-            def compute_descriptor(direction, i, j, positions):
-                atoms_temp = deepcopy(atoms)
-                atoms_temp.set_positions(positions)
-                desc_temp = model.featurize_structure(atoms.positions,
-                                                      atoms.get_chemical_symbols(),
-                                                      cell=np.array([atoms.cell[i] for i in range(3)]),
-                                                      pbc=atoms.pbc)
-                return direction, i, j, desc_temp
+    n_atoms = len(atoms)
+    n_mask = len(atoms_mask)
+    desc = features_batch[:n_atoms]  # (Natoms, Ndescriptor)
 
-            all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
-                delayed(compute_descriptor)(direction, i, j, positions)
-                for (direction, i, j, positions) in perturbations
-            )
+    n_atoms, n_features = desc.shape
 
-    else:
-        desc = model.featurize_structure(atoms.positions,
-                                         atoms.get_chemical_symbols())
-        n_atoms, D = desc.shape
+    # Pre-allocate gradient array
+    grad = torch.empty((n_atoms, n_atoms, 3, n_features), dtype=dtype)
 
-        # Pre-allocate gradient array
-        grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+    for atom_idx, i in enumerate(atoms_mask):
+        # 6 perturbations: +x, -x, +y, -y, +z, -z
+        desc_f_x = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 0): n_atoms + n_atoms * (atom_idx * 6 + 1)]
+        desc_b_x = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 1): n_atoms + n_atoms * (atom_idx * 6 + 2)]
+        desc_f_y = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 2): n_atoms + n_atoms * (atom_idx * 6 + 3)]
+        desc_b_y = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 3): n_atoms + n_atoms * (atom_idx * 6 + 4)]
+        desc_f_z = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 4): n_atoms + n_atoms * (atom_idx * 6 + 5)]
+        desc_b_z = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 5): n_atoms + n_atoms * (atom_idx * 6 + 6)]
 
-        # Get original positions once
-        original_positions = atoms.get_positions()
+        # Central difference: shape = (Natoms, Ndescriptor)
+        d_dx = (desc_f_x - desc_b_x) / (2 * delta)
+        d_dy = (desc_f_y - desc_b_y) / (2 * delta)
+        d_dz = (desc_f_z - desc_b_z) / (2 * delta)
 
-        # Pre-create all perturbations
-        perturbations = []
-        for i in atoms_mask:
-            for j in range(3):  # x, y, z
-                # Forward perturbation
-                pos_f = original_positions.copy()
-                pos_f[i, j] += delta
-                perturbations.append(('f', i, j, pos_f))
+        # (Natoms, 3, Ndescriptor) → assign to dfp_dr[idx, :, i, :, :]
+        grad[:, i, 0, :] = d_dx
+        grad[:, i, 1, :] = d_dy
+        grad[:, i, 2, :] = d_dz
 
-                # Backward perturbation
-                pos_b = original_positions.copy()
-                pos_b[i, j] -= delta
-                perturbations.append(('b', i, j, pos_b))
+    # n_atoms -> n_reduced_atoms
+    desc = desc[atoms_mask, :]
+    grad = grad[atoms_mask, :, :, :]
+    grad = grad[:, atoms_mask, :, :]
 
-        if n_jobs == 1:
-            # Compute all descriptors
-            all_descriptors = []
-            for direction, i, j, positions in perturbations:
-                atoms_temp = deepcopy(atoms)
-                atoms_temp.set_positions(positions)
-                desc_temp = model.featurize_structure(atoms.positions,
-                                                      atoms.get_chemical_symbols())
-                all_descriptors.append((direction, i, j, desc_temp))
-        else:
-            # --- Parallelize descriptor computation ---
-            def compute_descriptor(direction, i, j, positions):
-                atoms_temp = deepcopy(atoms)
-                atoms_temp.set_positions(positions)
-                desc_temp = model.featurize_structure(atoms.positions,
-                                                      atoms.get_chemical_symbols())
-                return direction, i, j, desc_temp
+    return desc, grad
 
-            all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
-                delayed(compute_descriptor)(direction, i, j, positions)
-                for (direction, i, j, positions) in perturbations
-            )
 
-    # Reorganize and compute gradients
-    desc_dict = {}
-    for direction, i, j, desc_temp in all_descriptors:
-        key = (i, j)
-        if key not in desc_dict:
-            desc_dict[key] = {}
-        desc_dict[key][direction] = desc_temp[:, :]
+def chebyshev_descriptor_gradient_periodic(atoms, model, atoms_mask, delta=1e-4, dtype=torch.float64):
+    """
+    Args:
+        atoms (ase.Atoms):
+        model: Chebyshev calculator
+        atoms_mask: Atom index to keep
+        delta (float): displacement (Ang)
+        dtype (dtype):
 
-    # Compute central differences
-    for (i, j), descs in desc_dict.items():
-        grad[:, i, j, :] = (descs['f'] - descs['b']) / (2 * delta)
+    Returns:
+        desc (torch.Tensor): (n_reduced_atoms, descriptor_dim)
+        grad (torch.Tensor): (n_reduced_atoms, n_reduced_atoms, 3, descriptor_dim)
+    """
+    # Get original descriptor
+    batch_positions = [(torch.tensor(atoms.positions, dtype=dtype))]
+    original_positions = atoms.get_positions()
+
+    for i in atoms_mask:
+        for j in range(3):
+            # Forward perturbation
+            pos_f = original_positions.copy()
+            pos_f[i, j] += delta
+            batch_positions.append(torch.tensor(pos_f, dtype=dtype))
+
+            # Backward perturbation
+            pos_b = original_positions.copy()
+            pos_b[i, j] -= delta
+            batch_positions.append(torch.tensor(pos_b, dtype=dtype))
+
+    batch_species = [atoms.get_chemical_symbols() for i in range(len(batch_positions))]
+    batch_cells = [torch.tensor(np.array([atoms.cell[i] for i in range(3)])) for i in range(len(batch_positions))]
+    batch_pbc = [torch.tensor(atoms.pbc) for i in range(len(batch_positions))]
+
+    # (Natoms * (6 * Nmaks + 1), Ndescriptor)
+    features_batch, batch_indices = model(batch_positions,
+                                          batch_species,
+                                          batch_cells=batch_cells,
+                                          batch_pbc=batch_pbc)
+
+    n_atoms = len(atoms)
+    n_mask = len(atoms_mask)
+    desc = features_batch[:n_atoms]
+
+    n_atoms, n_features = desc.shape
+
+    # Pre-allocate gradient array
+    grad = torch.empty((n_atoms, n_atoms, 3, n_features), dtype=dtype)
+
+    for atom_idx, i in enumerate(atoms_mask):
+        # 6 perturbations: +x, -x, +y, -y, +z, -z
+        desc_f_x = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 0): n_atoms + n_atoms * (atom_idx * 6 + 1)]
+        desc_b_x = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 1): n_atoms + n_atoms * (atom_idx * 6 + 2)]
+        desc_f_y = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 2): n_atoms + n_atoms * (atom_idx * 6 + 3)]
+        desc_b_y = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 3): n_atoms + n_atoms * (atom_idx * 6 + 4)]
+        desc_f_z = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 4): n_atoms + n_atoms * (atom_idx * 6 + 5)]
+        desc_b_z = features_batch[n_atoms + n_atoms * (atom_idx * 6 + 5): n_atoms + n_atoms * (atom_idx * 6 + 6)]
+
+        # Central difference: shape = (Natoms, Ndescriptor)
+        d_dx = (desc_f_x - desc_b_x) / (2 * delta)
+        d_dy = (desc_f_y - desc_b_y) / (2 * delta)
+        d_dz = (desc_f_z - desc_b_z) / (2 * delta)
+
+        # (Natoms, 3, Ndescriptor) → assign to dfp_dr[idx, :, i, :, :]
+        grad[:, i, 0, :] = d_dx
+        grad[:, i, 1, :] = d_dy
+        grad[:, i, 2, :] = d_dz
 
     # n_atoms -> n_reduced_atoms
     desc = desc[atoms_mask, :]
@@ -304,6 +312,7 @@ class GaussianProcess(object):
                 print("[2] L. Himanen, A. S Foster et al., Comput. Phys. Commun. 247 (2020) 106949. \n")
                 print("SOAP parameter:")
                 print(self.soap_param)
+                print("\n")
             except ImportError:
                 raise ImportError(
                     "The 'dscribe' package is required for using SOAP descriptors.\n"
@@ -330,6 +339,7 @@ class GaussianProcess(object):
                     print("[2] I. Batatia, G. Csányi et al., arXiv:2401.00096 (2023). \n")
                     print("MACE parameter:")
                     print(self.mace_param)
+                    print("\n")
                 except ImportError:
                     raise ImportError(
                         "The 'joblib' and 'mace' packages are required for using pre-trained MACE descriptors.\n"
@@ -340,42 +350,26 @@ class GaussianProcess(object):
 
                 # Check and set device
                 if torch.cuda.is_available():
+                    print(
+                        "[Note] There is available CUDA device, and it will be used for MACE descriptor computation.\n")
                     mace_device = "cuda:0"
                 else:
                     mace_device = "cpu"
-                    print("[Warning] CUDA device not available. MACE descriptor computation may be slow on CPU.")
+                    print("[Warning] CUDA device not available. MACE descriptor computation may be slow on CPU.\n")
 
                 self.mace = mace_mp(model=self.mace_param.get('model'),
                                     device=mace_device)
 
-            elif self.descriptor == 'chebyshev':
-                try:
-                    from aenet.torch_featurize import ChebyshevDescriptor
-                    print("You are using Chebyshev descriptor:")
-                    print("N. Artrith, A. Urban, and G. Ceder, Phys. Rev. B 96 (2017) 014112. \n")
-                    print("Chebshev parameter:")
-                    print(self.cheb_param)
-                except ImportError:
-                    raise ImportError(
-                        "The 'aenet-python' package is required for using Chebyshev descriptors.\n"
-                        "Please install it by running:\n\n"
-                        "    git clone https://github.com/atomisticnet/aenet-python.git\n"
-                        "    cd ./aenet-python/\n"
-                        "    pip install . --user\n"
-                    )
-
-                self.cheb = ChebyshevDescriptor(species=set(self.species),
-                                                rad_order=self.cheb_param.get("rad_order"),  # Radial polynomial order
-                                                rad_cutoff=self.cheb_param.get("rad_cutoff"),
-                                                # Radial cutoff (Angstroms)
-                                                ang_order=self.cheb_param.get("ang_order"),  # Angular polynomial order
-                                                ang_cutoff=self.cheb_param.get("ang_cutoff")
-                                                # Angular cutoff (Angstroms)
-                                                )
-
             else:
                 try:
                     from mace.calculators import mace_off
+                    print("You are using pre-trained MACE descriptor:")
+                    print(
+                        "[1] I. Batatia, D. P Kovacs, G. Simm, C. Ortner, and G. Csányi, Adv. Neural Inf. Process. Syst. 35 (2022) 11423.")
+                    print("[2] I. Batatia, G. Csányi et al., arXiv:2401.00096 (2023). \n")
+                    print("MACE parameter:")
+                    print(self.mace_param)
+                    print("\n")
                 except ImportError:
                     raise ImportError(
                         "The 'joblib' and 'mace' packages are required for using pre-trained MACE descriptors.\n"
@@ -384,11 +378,53 @@ class GaussianProcess(object):
                         "    pip install mace-torch\n"
                     )
 
+                # Check and set device
+                if torch.cuda.is_available():
+                    print(
+                        "[Note] There is available CUDA device, and it will be used for MACE descriptor computation.\n")
+                    mace_device = "cuda:0"
+                else:
+                    mace_device = "cpu"
+                    print("[Warning] CUDA device not available. MACE descriptor computation may be slow on CPU.\n")
+
                 self.mace = mace_off(model=self.mace_param.get('model'),
-                                     device=self.device)
+                                     device=mace_device)
+
+        elif self.descriptor == 'chebyshev':
+            try:
+                from aenet.torch_featurize import ChebyshevDescriptor, BatchedFeaturizer
+                print("You are using Chebyshev descriptor:")
+                print("N. Artrith, A. Urban, and G. Ceder, Phys. Rev. B 96 (2017) 014112. \n")
+                print("Chebshev parameter:")
+                print(self.cheb_param)
+                print("\n")
+            except ImportError:
+                raise ImportError(
+                    "The 'aenet-python' package is required for using Chebyshev descriptors.\n"
+                    "Please install it by running:\n\n"
+                    "    git clone https://github.com/atomisticnet/aenet-python.git\n"
+                    "    cd ./aenet-python/\n"
+                    "    pip install . --user\n"
+                )
+
+            # Check and set device
+            if torch.cuda.is_available():
+                print("[Note] There is available CUDA device, and it will be used for Chebyshev descriptor computation.\n")
+                cheb_device = "cuda:0"
+            else:
+                cheb_device = "cpu"
+                # print("[Warning] CUDA device not available. Chebyshev descriptor computation may be slow on CPU.")
+
+            self.chebyshev = ChebyshevDescriptor(species=set(self.species),
+                                                 rad_order=self.cheb_param.get("rad_order"),  # Radial polynomial order
+                                                 rad_cutoff=self.cheb_param.get("rad_cutoff"),  # Radial cutoff (Ang)
+                                                 ang_order=self.cheb_param.get("ang_order"),  # Angular polynomial order
+                                                 ang_cutoff=self.cheb_param.get("ang_cutoff"),  # Angular cutoff (Ang)
+                                                 device=cheb_device)
+            self.chebyshev_batch = BatchedFeaturizer(self.chebyshev)
 
         self.atoms_xyz_mask = atoms_mask.to(self.device)
-        self.Nmask = self.atoms_xyz_mask.shape[0]  # 3 * Natoms or 3 * Nreduced_atoms
+        self.Nmask_xyz = self.atoms_xyz_mask.shape[0]  # 3 * Natoms or 3 * Nreduced_atoms
         self.atoms_mask = (self.atoms_xyz_mask[self.atoms_xyz_mask % 3 == 0] // 3).to(self.device)
 
         # train_fp: (Ndata, Ncenter, Ndescriptor)
@@ -406,7 +442,7 @@ class GaussianProcess(object):
         Nreduced_atoms = self.atoms_mask.shape[0]
         self.dY = self.dY.view(self.Ntrain, Nreduced_atoms, 3)
 
-        self.model_vector = torch.empty((self.Ntrain * (1 + self.Nmask),), dtype=self.torch_data_type,
+        self.model_vector = torch.empty((self.Ntrain * (1 + self.Nmask_xyz),), dtype=self.torch_data_type,
                                         device=self.device)
 
         if prior is None:
@@ -436,7 +472,7 @@ class GaussianProcess(object):
         self.kernel = FPKernel(species=self.species,
                                pbc=self.pbc,
                                Natom=self.Natom,
-                               Nmask=self.Nmask,
+                               Nmask=self.Nmask_xyz,
                                kerneltype=self.kerneltype,
                                data_type=self.data_type,
                                device=self.device)
@@ -512,14 +548,23 @@ class GaussianProcess(object):
             fp = []
             dfp_dr = []
             for image in images:
-                fp__, dfp_dr__ = chebyshev_descriptor_gradient(image,
-                                                               self.cheb,
-                                                               atoms_mask=self.atoms_mask,
-                                                               delta=self.cheb_param.get("delta"),
-                                                               n_jobs=self.cheb_param.get("n_jobs"),
-                                                               dtype=self.data_type)
-                fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
-                dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
+                if np.any(image.pbc):
+                    fp__, dfp_dr__ = chebyshev_descriptor_gradient_periodic(image,
+                                                                            self.chebyshev_batch,
+                                                                            atoms_mask=self.atoms_mask,
+                                                                            delta=self.mace_param.get("delta"),
+                                                                            dtype=self.torch_data_type)
+                    fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
+                    dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
+
+                else:
+                    fp__, dfp_dr__ = chebyshev_descriptor_gradient(image,
+                                                                   self.chebyshev_batch,
+                                                                   atoms_mask=self.atoms_mask,
+                                                                   delta=self.mace_param.get("delta"),
+                                                                   dtype=self.torch_data_type)
+                    fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
+                    dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
 
             fp = torch.stack(fp).to(dtype=self.torch_data_type,
                                     device=self.device)  # (Ndata, Nreduced_atoms, Ndescriptor)
@@ -574,16 +619,23 @@ class GaussianProcess(object):
                                   device=self.device)  # (Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
 
         elif self.descriptor == 'chebyshev':
-            fp, dfp_dr = chebyshev_descriptor_gradient(image,
-                                                       self.cheb,
-                                                       atoms_mask=self.atoms_mask,
-                                                       delta=self.cheb_param.get("delta"),
-                                                       n_jobs=self.cheb_param.get("n_jobs"),
-                                                       dtype=self.data_type)
+            if np.any(image.pbc):
+                fp, dfp_dr = chebyshev_descriptor_gradient_periodic(image,
+                                                                    self.chebyshev_batch,
+                                                                    atoms_mask=self.atoms_mask,
+                                                                    delta=self.mace_param.get("delta"),
+                                                                    dtype=self.torch_data_type)
 
-            fp = torch.tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Nreduced_atoms, Ndescriptor)
-            dfp_dr = torch.tensor(dfp_dr, dtype=self.torch_data_type,
-                                  device=self.device)  # (Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
+            else:
+                fp, dfp_dr = chebyshev_descriptor_gradient(image,
+                                                           self.chebyshev_batch,
+                                                           atoms_mask=self.atoms_mask,
+                                                           delta=self.mace_param.get("delta"),
+                                                           dtype=self.torch_data_type)
+
+            fp = fp.to(dtype=self.torch_data_type, device=self.device)  # (Nreduced_atoms, Ndescriptor)
+            dfp_dr = dfp_dr.to(dtype=self.torch_data_type,
+                               device=self.device)  # (Nreduced_atoms, Nreduced_atoms, 3, Ndescriptor)
 
         else:
             fp = torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(
@@ -607,7 +659,7 @@ class GaussianProcess(object):
         a = torch.full((self.Ntrain, 1), self.hyper_params['noise'] * self.hyper_params['noisefactor'],
                        dtype=self.torch_data_type, device=self.device)
         noise_val = self.hyper_params['noise']
-        b = noise_val.expand(self.Ntrain, self.Nmask)
+        b = noise_val.expand(self.Ntrain, self.Nmask_xyz)
 
         # reg = torch.diag(torch.cat((a, b), 1).flatten() ** 2)
         diagonal_values = torch.cat((a, b), 1).flatten() ** 2
