@@ -1,5 +1,4 @@
 import torch
-import gc
 
 from aenet_gpr.src.pytorch_kerneltypes import SquaredExp
 from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch
@@ -39,6 +38,7 @@ class FPKernel(BaseKernelType):
                  species,
                  pbc,
                  Natom,
+                 Nmask,
                  kerneltype='sqexp',
                  params=None,
                  data_type='float64',
@@ -61,6 +61,7 @@ class FPKernel(BaseKernelType):
         self.species = species
         self.pbc = pbc
         self.Natom = Natom
+        self.Nmask = Nmask
 
         if params is None:
             params = {}
@@ -107,8 +108,8 @@ class FPKernel(BaseKernelType):
         #  ...
         #  [K(dX1Nz,X21), K(dX1Nz,X22), ..., K(dX1Nz,X2N), K(dX1Nz,dX21x), K(dX1Nz,dX21y), ..., K(dX1Nz,dX2Nz)]
 
-        K_X1X2 = torch.empty(((1 + 3 * self.Natom) * X1.shape[0],
-                              (1 + 3 * self.Natom) * X2.shape[0]), dtype=self.torch_data_type, device=self.device)
+        K_X1X2 = torch.empty(((1 + self.Nmask) * X1.shape[0],
+                              (1 + self.Nmask) * X2.shape[0]), dtype=self.torch_data_type, device=self.device)
 
         # Expand value fingerprint
         X1_expanded = X1[:, None, :, None, :]  # [Ndata1, 1, Ncenter, 1, Nfeature]
@@ -131,6 +132,7 @@ class FPKernel(BaseKernelType):
         """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), :X2.shape[0]] kernel between fp1_deriv and fp2"""
         # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
         dX1_reshaped = dX1.flatten(start_dim=2, end_dim=3)
+        del dX1
         # intermediate_result = torch.einsum('xycnf,xcdf->xycnd', X1__outer_minus__X2, dX1_reshaped)
         # intermediate_result = torch.matmul(X1__outer_minus__X2, dX1_reshaped.transpose(-1, -2))
 
@@ -142,15 +144,14 @@ class FPKernel(BaseKernelType):
         # K_X1X2[X1.shape[0]:, :X2.shape[0]] = kernel.permute(0, 2, 1).reshape(X1.shape[0] * 3 * self.Natom, X2.shape[0])
         # K_X1X2[X1.shape[0]:, :X2.shape[0]].copy_(
         #     kernel.permute(0, 2, 1).contiguous().view(X1.shape[0] * 3 * self.Natom, X2.shape[0]))
-        K_X1X2[X1.shape[0]:, :X2.shape[0]].copy_(torch.einsum('xycn,xycnd->xyd', __k_X1X2 / self.scale ** 2,
-                                                              torch.einsum('xycnf,xcdf->xycnd', X1__outer_minus__X2,
-                                                                           dX1_reshaped)).permute(0, 2,
-                                                                                                  1).contiguous().view(
-            X1.shape[0] * 3 * self.Natom, X2.shape[0]))
+        tmp1 = torch.einsum('xycnf,xcdf->xycnd', X1__outer_minus__X2, dX1_reshaped)
+        tmp1 = torch.einsum('xycn,xycnd->xyd', __k_X1X2 / self.scale ** 2, tmp1)
+        K_X1X2[X1.shape[0]:, :X2.shape[0]].copy_(tmp1.permute(0, 2, 1).contiguous().view(X1.shape[0] * self.Nmask, X2.shape[0]))
 
         """K_X1X2[:X1.shape[0], X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] kernel between fp1 and fp2_deriv"""
         # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
         dX2_reshaped = dX2.flatten(start_dim=2, end_dim=3)
+        del dX2
         # intermediate_result = torch.einsum('xycnf,yndf->xycnd', X2__outer_minus__X1, dX2_reshaped)
         # intermediate_result = torch.matmul(X2__outer_minus__X1, dX2_reshaped.transpose(-1, -2))
 
@@ -161,10 +162,9 @@ class FPKernel(BaseKernelType):
         # K_X1X2[:X1.shape[0], X2.shape[0]:] = kernel.reshape(X1.shape[0], X2.shape[0] * 3 * self.Natom)
         # K_X1X2[:X1.shape[0], X2.shape[0]:].copy_(
         #     kernel.view(X1.shape[0], X2.shape[0] * 3 * self.Natom))
-        K_X1X2[:X1.shape[0], X2.shape[0]:].copy_(torch.einsum('xycn,xycnd->xyd', __k_X2X1 / self.scale ** 2,
-                                                              torch.einsum('xycnf,yndf->xycnd', X2__outer_minus__X1,
-                                                                           dX2_reshaped)).view(X1.shape[0], X2.shape[
-            0] * 3 * self.Natom))
+        tmp1 = torch.einsum('xycnf,yndf->xycnd', X2__outer_minus__X1, dX2_reshaped)
+        tmp1 = torch.einsum('xycn,xycnd->xyd', __k_X2X1 / self.scale ** 2, tmp1)
+        K_X1X2[:X1.shape[0], X2.shape[0]:].copy_(tmp1.contiguous().view(X1.shape[0], X2.shape[0] * self.Nmask))
 
         """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] 
         kernel between fp1_deriv and fp2_deriv"""
@@ -226,26 +226,22 @@ class FPKernel(BaseKernelType):
         #                                                  X2.shape[0] * 3 * self.Natom))
 
         # intermediate_result = torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1, DdD_dr2)
-        K_X1X2[X1.shape[0]:, X2.shape[0]:].copy_(torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 4,
-                                                              torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1,
-                                                                           DdD_dr2)).permute(0, 2, 1,
-                                                                                             3).contiguous().view(
-            X1.shape[0] * 3 * self.Natom, X2.shape[0] * 3 * self.Natom))
+        tmp1 = torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1,  DdD_dr2)
+        tmp1 = torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 4, tmp1)
+        K_X1X2[X1.shape[0]:, X2.shape[0]:].copy_(tmp1.permute(0, 2, 1, 3).contiguous().view(X1.shape[0] * self.Nmask, X2.shape[0] * self.Nmask))
 
         # intermediate_result = torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped, dX2_reshaped)
-        K_X1X2[X1.shape[0]:, X2.shape[0]:].add_(torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 2,
-                                                             torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped,
-                                                                          dX2_reshaped)).permute(0, 2, 1,
-                                                                                                 3).contiguous().view(
-            X1.shape[0] * 3 * self.Natom, X2.shape[0] * 3 * self.Natom))
+        tmp1 = torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped, dX2_reshaped)
+        tmp1 = torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 2, tmp1)
+        K_X1X2[X1.shape[0]:, X2.shape[0]:].add_(tmp1.permute(0, 2, 1, 3).contiguous().view(X1.shape[0] * self.Nmask, X2.shape[0] * self.Nmask))
 
-        del X1_expanded, X2_expanded, __k_X1X2, __k_X2X1, X1__outer_minus__X2, X2__outer_minus__X1, \
-            DdD_dr1, DdD_dr2, dX1_reshaped, dX2_reshaped
-        gc.collect()
+        # del X1_expanded, X2_expanded, __k_X1X2, __k_X2X1, X1__outer_minus__X2, X2__outer_minus__X1, \
+        #     DdD_dr1, DdD_dr2, dX1_reshaped, dX2_reshaped
+        # gc.collect()
 
         return K_X1X2
 
-    def kernel_per_data(self, X1=None, dX1=None, X2=None, dX2=None, iter=0):
+    def kernel_per_data(self, X1=None, dX1=None, X2=None, dX2=None):
         '''
         Return a full kernel matrix between two structural fingerprints, 'X1' and 'X2'.
         X1, X2 = [Ncenter, Nfeature]
@@ -263,8 +259,8 @@ class FPKernel(BaseKernelType):
         k = self.weight ** 2 * torch.exp(-torch.linalg.norm(X1_outer_minus_X2, dim=2) ** 2 / (2 * self.scale ** 2))
 
         # Create empty kernel
-        kernel = torch.empty((1 + 3 * self.Natom,
-                              1 + 3 * self.Natom), dtype=self.torch_data_type, device=self.device)
+        kernel = torch.empty((1 + self.Nmask,
+                              1 + self.Nmask), dtype=self.torch_data_type, device=self.device)
 
         # Evaluate the first element of kernel
         kernel[0, 0] = torch.sum(k)
@@ -273,7 +269,7 @@ class FPKernel(BaseKernelType):
         intermediate_result = torch.einsum('cnf,c...af->cna',
                                            X1_outer_minus_X2,
                                            dX1.reshape(X1.shape[0],
-                                                       3 * self.Natom,
+                                                       self.Nmask,
                                                        Nfeature))
         kernel[1:, 0] = torch.einsum('cn,cna->a',
                                      k / self.scale ** 2,
@@ -283,7 +279,7 @@ class FPKernel(BaseKernelType):
         intermediate_result = torch.einsum('cnf,...naf->cna',
                                            X2_outer_minus_X1,
                                            dX2.reshape(X2.shape[0],
-                                                       3 * self.Natom,
+                                                       self.Nmask,
                                                        Nfeature))
         kernel[0, 1:] = torch.einsum('cn,cna->a',
                                      k / self.scale ** 2,
@@ -294,12 +290,12 @@ class FPKernel(BaseKernelType):
         DdD_dr1 = torch.einsum('cnf,c...bf->cnb',
                                X1_outer_minus_X2,
                                dX1.reshape(X1.shape[0],
-                                           3 * self.Natom,
+                                           self.Nmask,
                                            Nfeature))
         DdD_dr2 = torch.einsum('cnf,...ndf->cnd',
                                X2_outer_minus_X1,
                                dX2.reshape(X2.shape[0],
-                                           3 * self.Natom,
+                                           self.Nmask,
                                            Nfeature))
 
         # [Ncenter, Natom * 3] * [Ncenter, Natom * 3]
@@ -314,20 +310,20 @@ class FPKernel(BaseKernelType):
         # -> [Ncenter, Natom * 3, Natom * 3]
         intermediate_result = torch.einsum('cbf,ndf->cnbd',
                                            dX1.reshape(X1.shape[0],
-                                                       3 * self.Natom,
+                                                       self.Nmask,
                                                        Nfeature),
                                            dX2.reshape(X2.shape[0],
-                                                       3 * self.Natom,
+                                                       self.Nmask,
                                                        Nfeature))
         C1 = torch.einsum('cn,cnbd->bd',
                           k / self.scale ** 2,
                           intermediate_result)
         kernel[1:, 1:] = C0 + C1
 
-        if iter % 10 == 9:
-            del X1_expanded, X2_expanded, X1_outer_minus_X2, X2_outer_minus_X1, k, intermediate_result, \
-                DdD_dr1, DdD_dr2, C0, C1
-            gc.collect()
+        # if iter % 10 == 9:
+        #     del X1_expanded, X2_expanded, X1_outer_minus_X2, X2_outer_minus_X1, k, intermediate_result, \
+        #         DdD_dr1, DdD_dr2, C0, C1
+        #     gc.collect()
 
         return kernel
 
@@ -343,16 +339,16 @@ class FPKernel(BaseKernelType):
         X_N_batch = get_N_batch(Ndata, batch_size)
         X_indexes = get_batch_indexes_N_batch(Ndata, X_N_batch)
 
-        K_XX = torch.empty((Ndata * (1 + 3 * self.Natom),
-                            Ndata * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
+        K_XX = torch.empty((Ndata * (1 + self.Nmask),
+                            Ndata * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
 
         for i in range(0, X_N_batch):
             fp_i = fp[X_indexes[i][0]:X_indexes[i][1], :, :]
             dfp_dr_i = dfp_dr[X_indexes[i][0]:X_indexes[i][1], :, :, :, :]
 
             row_indexes = torch.arange(X_indexes[i][0], X_indexes[i][1])
-            row_deriv_indexes = torch.arange(Ndata + X_indexes[i][0] * 3 * self.Natom,
-                                             Ndata + X_indexes[i][1] * 3 * self.Natom)
+            row_deriv_indexes = torch.arange(Ndata + X_indexes[i][0] * self.Nmask,
+                                             Ndata + X_indexes[i][1] * self.Nmask)
             selected_rows = torch.cat([row_indexes, row_deriv_indexes])
 
             K_XX[selected_rows[:, None], selected_rows] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
@@ -363,8 +359,8 @@ class FPKernel(BaseKernelType):
                 dfp_dr_j = dfp_dr[X_indexes[j][0]:X_indexes[j][1], :, :, :, :]
 
                 col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
-                col_deriv_indexes = torch.arange(Ndata + X_indexes[j][0] * 3 * self.Natom,
-                                                 Ndata + X_indexes[j][1] * 3 * self.Natom)
+                col_deriv_indexes = torch.arange(Ndata + X_indexes[j][0] * self.Nmask,
+                                                 Ndata + X_indexes[j][1] * self.Nmask)
                 selected_cols = torch.cat([col_indexes, col_deriv_indexes])
 
                 K_XX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
@@ -377,32 +373,30 @@ class FPKernel(BaseKernelType):
 
         Ndata = fp.shape[0]
 
-        K_XX = torch.empty((Ndata * (1 + 3 * self.Natom),
-                            Ndata * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
+        K_XX = torch.empty((Ndata * (1 + self.Nmask),
+                            Ndata * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
 
         for i in range(0, Ndata):
             fp_i = fp[i, :, :]
             dfp_dr_i = dfp_dr[i, :, :, :, :]
 
             row_index = torch.tensor([i])
-            row_deriv_index = torch.arange(Ndata + i * 3 * self.Natom, Ndata + (i + 1) * 3 * self.Natom)
+            row_deriv_index = torch.arange(Ndata + i * self.Nmask, Ndata + (i + 1) * self.Nmask)
             selected_rows = torch.cat([row_index, row_deriv_index])
 
             K_XX[selected_rows[:, None], selected_rows] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
-                                                                               X2=fp_i, dX2=dfp_dr_i,
-                                                                               iter=i)
+                                                                               X2=fp_i, dX2=dfp_dr_i)
 
             for j in range(i + 1, Ndata):
                 fp_j = fp[j, :, :]
                 dfp_dr_j = dfp_dr[j, :, :, :, :]
 
                 col_index = torch.tensor([j])
-                col_deriv_index = torch.arange(Ndata + j * 3 * self.Natom, Ndata + (j + 1) * 3 * self.Natom)
+                col_deriv_index = torch.arange(Ndata + j * self.Nmask, Ndata + (j + 1) * self.Nmask)
                 selected_cols = torch.cat([col_index, col_deriv_index])
 
                 K_XX[selected_rows[:, None], selected_cols] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
-                                                                                   X2=fp_j, dX2=dfp_dr_j,
-                                                                                   iter=j)
+                                                                                   X2=fp_j, dX2=dfp_dr_j)
                 K_XX[selected_cols[:, None], selected_rows] = K_XX[selected_rows[:, None], selected_cols].T
 
         return K_XX
@@ -432,16 +426,16 @@ class FPKernel(BaseKernelType):
         X_N_batch = get_N_batch(Ntrain, batch_size)
         X_indexes = get_batch_indexes_N_batch(Ntrain, X_N_batch)
 
-        K_xX = torch.empty((Ntest * (1 + 3 * self.Natom),
-                            Ntrain * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
+        K_xX = torch.empty((Ntest * (1 + self.Nmask),
+                            Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
 
         for i in range(0, x_N_batch):
             fp_1_i = fp_1[x_indexes[i][0]:x_indexes[i][1], :, :]
             dfp_dr_1_i = dfp_dr_1[x_indexes[i][0]:x_indexes[i][1], :, :, :, :]
 
             row_indexes = torch.arange(x_indexes[i][0], x_indexes[i][1])
-            row_deriv_indexes = torch.arange(Ntest + x_indexes[i][0] * 3 * self.Natom,
-                                             Ntest + x_indexes[i][1] * 3 * self.Natom)
+            row_deriv_indexes = torch.arange(Ntest + x_indexes[i][0] * self.Nmask,
+                                             Ntest + x_indexes[i][1] * self.Nmask)
             selected_rows = torch.cat([row_indexes, row_deriv_indexes])
 
             for j in range(0, X_N_batch):
@@ -449,8 +443,8 @@ class FPKernel(BaseKernelType):
                 dfp_dr_2_j = dfp_dr_2[X_indexes[j][0]:X_indexes[j][1], :, :, :, :]
 
                 col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
-                col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * 3 * self.Natom,
-                                                 Ntrain + X_indexes[j][1] * 3 * self.Natom)
+                col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * self.Nmask,
+                                                 Ntrain + X_indexes[j][1] * self.Nmask)
                 selected_cols = torch.cat([col_indexes, col_deriv_indexes])
 
                 K_xX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_1_i, dX1=dfp_dr_1_i,
@@ -472,15 +466,15 @@ class FPKernel(BaseKernelType):
         Ntrain = fp_2.shape[0]
 
         # if dx is not None and dX is not None:
-        K_xX = torch.empty((Ntest * (1 + 3 * self.Natom),
-                            Ntrain * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
+        K_xX = torch.empty((Ntest * (1 + self.Nmask),
+                            Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
 
         for i in range(0, Ntest):
             fp_1_i = fp_1[i, :, :]
             dfp_dr_1_i = dfp_dr_1[i, :, :, :, :]
 
             row_index = torch.tensor([i])
-            row_deriv_index = torch.arange(Ntest + i * 3 * self.Natom, Ntest + (i + 1) * 3 * self.Natom)
+            row_deriv_index = torch.arange(Ntest + i * self.Nmask, Ntest + (i + 1) * self.Nmask)
             selected_rows = torch.cat([row_index, row_deriv_index])
 
             for j in range(0, Ntrain):
@@ -488,12 +482,11 @@ class FPKernel(BaseKernelType):
                 dfp_dr_2_j = dfp_dr_2[j, :, :, :, :]
 
                 col_index = torch.tensor([j])
-                col_deriv_index = torch.arange(Ntrain + j * 3 * self.Natom, Ntrain + (j + 1) * 3 * self.Natom)
+                col_deriv_index = torch.arange(Ntrain + j * self.Nmask, Ntrain + (j + 1) * self.Nmask)
                 selected_cols = torch.cat([col_index, col_deriv_index])
 
                 K_xX[selected_rows[:, None], selected_cols] = self.kernel_per_data(X1=fp_1_i, dX1=dfp_dr_1_i,
-                                                                                   X2=fp_2_j, dX2=dfp_dr_2_j,
-                                                                                   iter=j)
+                                                                                   X2=fp_2_j, dX2=dfp_dr_2_j)
 
         return K_xX
 
@@ -515,16 +508,16 @@ class FPKernel(BaseKernelType):
         dfp_dr_1_i = dfp_dr_1_i.unsqueeze(0)
 
         # Ntrain * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
-        K_xX_i = torch.empty((1 * (1 + 3 * self.Natom),
-                              Ntrain * (1 + 3 * self.Natom)), dtype=self.torch_data_type, device=self.device)
+        K_xX_i = torch.empty((1 * (1 + self.Nmask),
+                              Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
 
         for j in range(0, X_N_batch):
             fp_2_j = fp_2[X_indexes[j][0]:X_indexes[j][1], :, :]
             dfp_dr_2_j = dfp_dr_2[X_indexes[j][0]:X_indexes[j][1], :, :, :, :]
 
             col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
-            col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * 3 * self.Natom,
-                                             Ntrain + X_indexes[j][1] * 3 * self.Natom)
+            col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * self.Nmask,
+                                             Ntrain + X_indexes[j][1] * self.Nmask)
             selected_cols = torch.cat([col_indexes, col_deriv_indexes])
 
             K_xX_i[:, selected_cols] = self.kernel_with_deriv(X1=fp_1_i, dX1=dfp_dr_1_i,
