@@ -15,7 +15,7 @@ except ModuleNotFoundError:
 
 from aenet_gpr.src import GPRCalculator
 from aenet_gpr.util import ReferenceData
-from aenet_gpr.tool import acquisition, prepare_neb_images, dump_observation, get_fmax
+from aenet_gpr.tool import acquisition, dump_observation, get_fmax
 from aenet_gpr.inout.input_parameter import InputParameters
 
 
@@ -30,6 +30,13 @@ def min_cartesian_dist(img, train_images):
     return best
 
 
+def is_duplicate(pos_new, existing_positions, atol=1e-4):
+    for pos_old in existing_positions:
+        if np.allclose(pos_new, pos_old, atol=atol):
+            return True
+    return False
+
+
 class AIDNEB:
 
     def __init__(self, start, end, input_param: InputParameters, model_calculator=None, calculator=None,
@@ -294,9 +301,35 @@ class AIDNEB:
         # print(f"r_max (threshold to prevent over-relaxation when training data is sparse): {self.rmax:.4f}")
         print('spring_constant: ', self.spring)
 
+    def save_neb_predictions_to_extxyz(self, predictions, filename):
+        """
+        Store NEB predictions (energy, force, unc_energy, unc_force)
+        into Atoms objects and write to extxyz file.
+
+        Parameters:
+            predictions (dict): Dictionary with 'energy', 'forces', 'unc_energy', 'unc_forces'
+            filename (str): Output file path (.extxyz)
+        """
+        out = []
+        for i, image in enumerate(self.images):
+            atoms = image.copy()
+
+            # Energies
+            atoms.info['energy'] = float(predictions['energy'][i])
+            atoms.info['unc_energy'] = float(predictions['unc_energy'][i])
+
+            # Forces
+            atoms.arrays['forces'] = np.linalg.norm(predictions['forces'][i], axis=1)
+            atoms.arrays['unc_forces'] = np.linalg.norm(predictions['unc_forces'][i], axis=1)
+
+            out.append(atoms)
+
+        # Write to extxyz file
+        io.write(filename, out, format='extxyz')
+
     def run(self,
             fmax=0.05,
-            unc_convergence=0.02,
+            unc_convergence=0.05,
             dt=0.1,
             ml_steps=150,
             optimizer="MDMin",
@@ -405,9 +438,24 @@ class AIDNEB:
             #     self.images = copy.deepcopy(self.initial_interpolation)
 
             # 1. Collect observations.
-            # This serves to use_previous_observations from a previous (and/or parallel) runs.
             train_images = io.read(trajectory_observations, ':')
 
+            """
+            # (N_candidate, N_atoms, 3)
+            train_positions = np.asarray([img.get_positions() for img in train_images], dtype=train_data.numpy_data_type)
+            # (N_train, N_atoms, 3)
+            original_positions = np.asarray(train_data.structure, dtype=train_data.numpy_data_type)
+
+            new_images = []
+            for img, pos in zip(train_images, train_positions):
+                if not is_duplicate(pos, original_positions):
+                    new_images.append(img)
+
+            if new_images:
+                pass
+                # train_data.update_train_data(new_images)
+            """
+
             train_data = ReferenceData(structure_files=train_images,
                                        file_format='ase',
                                        device=self.input_param.device,
@@ -422,9 +470,6 @@ class AIDNEB:
             if train_data.standardization:
                 train_data.standardize_energy_force(train_data.energy)
 
-            if check_ref_force:
-                self.input_param.filter = False
-
             # 2. Prepare a calculator.
             print('Training data size: ', len(train_images))
             print('Descriptor: ', self.input_param.descriptor)
@@ -527,13 +572,18 @@ class AIDNEB:
             F = F.reshape(nim, nat, 3)
             max_f_image = np.sqrt((F ** 2).sum(-1)).max().item()
 
-            predictions = get_neb_predictions(self.images[1:-1])
+            predictions = get_neb_predictions(self.images)
+            filename = f'gpr_neb_results_step{self.step:04d}.extxyz'
+            self.save_neb_predictions_to_extxyz(predictions=predictions, filename=filename)
+
             neb_pred_energy = predictions['energy']
-            neb_pred_uncertainty = predictions['uncertainty']
+            # neb_pred_unc_energy = predictions['unc_energy']
+            neb_pred_unc_forces = predictions['unc_forces']
 
             # 5. Print output.
             max_e = np.max(neb_pred_energy)
-            max_unc = np.max(neb_pred_uncertainty)
+            # max_unc = np.max(neb_pred_unc_energy)
+            max_unc = max(np.linalg.norm(f_unc, axis=1).max() for f_unc in neb_pred_unc_forces)
             self.max_unc_hist.append(max_unc)
 
             # Calculator of train_images is reference, while Calculator of self.images is GP
@@ -665,15 +715,27 @@ def make_neb(self, images_interpolation=None):
 def get_neb_predictions(images):
     neb_pred_energy = []
     neb_pred_forces = []
-    neb_pred_unc = []
+    neb_pred_unc_energy = []
+    neb_pred_unc_forces = []
 
     for i in images:
         neb_pred_energy.append(i.get_potential_energy())
         neb_pred_forces.append(i.get_forces())
-        unc = i.calc.results['uncertainty']
-        neb_pred_unc.append(unc)
 
-    predictions = {'energy': neb_pred_energy, 'forces': neb_pred_forces, 'uncertainty': neb_pred_unc}
+        unc_energy = i.calc.results['unc_energy']
+        neb_pred_unc_energy.append(unc_energy)
+
+        unc_forces = i.calc.results['unc_forces']
+        neb_pred_unc_forces.append(unc_forces)
+
+    neb_pred_unc_energy[0] = 0.0
+    neb_pred_unc_energy[-1] = 0.0
+
+    neb_pred_unc_forces[0] = np.zeros_like(neb_pred_unc_forces[1])
+    neb_pred_unc_forces[-1] = np.zeros_like(neb_pred_unc_forces[1])
+
+    predictions = {'energy': neb_pred_energy, 'forces': neb_pred_forces,
+                   'unc_energy': neb_pred_unc_energy, 'unc_forces': neb_pred_unc_forces}
 
     return predictions
 
